/*
 * yasapi_engine_common.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <stdint.h>

///////////////////////////////////////////////////////////////////////////////
#if ! defined (AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM) // [
#define AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM 0x80000000
#endif // ]

#if ! defined (AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY) // [
#define AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY 0x08000000
#endif // ]

///////////////////////////////////////////////////////////////////////////////
static int EngineCreateBase(Engine *pEngine, Player *pPlayer);
static int EngineCreateOptions(Engine *pEngine, Player *pPlayer);
static int EngineCreateWfx(Engine *pEngine, WAVEFORMATEXTENSIBLE *pWfx,
    DWORD samplerate, WORD numchannels, WORD bitspersamp);
static int EngineGetBufferDuration(Engine *pEngine);
static int EngineCreateRing(Engine *pEngine, DWORD dwChunkSizeSource);
#if defined (YASAPI_ENGINE_IS_PAUSE) // [
static int EngineIsPause(Engine *pEngine);
#endif // ]
static double EngineOutputTimeInt(Engine *pEngine);

///////////////////////////////////////////////////////////////////////////////
int EngineCreate(Engine *pEngine, Player *pPlayer, int samplerate,
    int numchannels, int bitspersamp)
{
  const EEngine eState=pEngine->eState;
  int nMaxLatency=0;
  WORD nChannelsTarget;
  WORD wBitsPerSampleTarget;
#if ! defined (YASAPI_ENGINE_ENUMERATOR) // [
  Wasapi *pWasapi;

  /////////////////////////////////////////////////////////////////////////////
  if (!(pWasapi=PlayerWasapi(pPlayer))) {
    DMESSAGE("accessing WASAPI");
    EngineDestroy(pEngine);
    goto exit;
  }

  pEngine->pDevice=pWasapi->selected.pDevice->pDevice;
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  ZeroMemory(&pEngine->params,sizeof pEngine->params);

  /////////////////////////////////////////////////////////////////////////////
  if (eState<eEngineBase&&EngineCreateBase(pEngine,pPlayer)<0) {
    DMESSAGE("creating engine base");
    EngineKill(pEngine,eState);
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  ZeroMemory(&pEngine->stats,sizeof pEngine->stats);
  pEngine->stats.uNumTracks=1u;
  ++pEngine->pPlayer->stats.uNumTracks;
  pEngine->params.msHistory=0.0;

  /////////////////////////////////////////////////////////////////////////////
  if (EngineCreateOptions(pEngine,pPlayer)<0) {
    DMESSAGE("creating options");
    EngineKill(pEngine,eState);
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (EngineCreateWfx(pEngine,&pEngine->wfx.source,samplerate,
      numchannels,bitspersamp)<0) {
    DVMESSAGE("unknown format %d/%d/%d",samplerate,numchannels,
        bitspersamp);
    EngineKill(pEngine,eState);
    goto exit;
  }

  if (1==numchannels&&pEngine->options.common.eMono2Stereo)
    nChannelsTarget=2;
  else
    nChannelsTarget=numchannels;

  switch (pEngine->options.device.eBitdepth) {
  case eBitdepth16:
    wBitsPerSampleTarget=16u;
    break;
  case eBitdepth24:
    wBitsPerSampleTarget=24u;
    break;
  case eBitdepth32:
    wBitsPerSampleTarget=32u;
    break;
  default:
    wBitsPerSampleTarget=bitspersamp;
    break;
  }

  if (EngineCreateWfx(pEngine,&pEngine->wfx.target,samplerate,
      nChannelsTarget,wBitsPerSampleTarget)<0) {
    DVMESSAGE("unknown format %d/%u/%u",samplerate,nChannelsTarget,
        wBitsPerSampleTarget);
    EngineKill(pEngine,eState);
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (EngineCreateAudioClient(pEngine,eEngineAudioClientInititial)<0) {
    DMESSAGE("creating audio client");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DWRITELN(EngineState(pEngine));
  DASSERT(eEngineAudioClientInititial==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
  GtkUpdateDialog(&pPlayer->gtk,&pPlayer->gtk.config);
  nMaxLatency=pEngine->options.device.xMaxLatencyAssumed+0.5;
exit:
  return nMaxLatency;
}

static int EngineCreateBase(Engine *pEngine, Player *pPlayer)
{
  const EEngine eState=pEngine->eState;
#if defined (YASAPI_ENGINE_ENUMERATOR) // [
  HRESULT hr;
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  pEngine->pPlayer=pPlayer;

  /////////////////////////////////////////////////////////////////////////////
  InitializeCriticalSection(&pEngine->cs);
  ++pEngine->eState;
  DASSERT(eEngineCriticalSection==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
  if (!(pEngine->timer.hEvent=CreateEventW(NULL,FALSE,FALSE,NULL))) {
    DMESSAGE("creating timer event");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineTimerEvent==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
  if (!(pEngine->timer.hTimer=CreateWaitableTimerW(NULL,FALSE,NULL))) {
    DMESSAGE("creating waitable timer");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineTimer==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
  if (!(pEngine->thread.hProducer=CreateEventW(NULL,FALSE,FALSE,NULL))) {
    DMESSAGE("creating producer event");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineEventProducer==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
  if (!(pEngine->thread.hConsumer=CreateEventW(NULL,FALSE,FALSE,NULL))) {
    DMESSAGE("creating consumer event");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineEventConsumer==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
  EnterCriticalSection(&pEngine->cs); // {
  pEngine->thread.pCommand=NULL;

  pEngine->thread.hThread=CreateThread(
    NULL,     // _In_opt_  LPSECURITY_ATTRIBUTES  lpThreadAttributes,
    0ul,      // _In_      SIZE_T                 dwStackSize,
    EngineThreadProc,
              // _In_      LPTHREAD_START_ROUTINE lpStartAddress,
    pEngine,  // _In_opt_  LPVOID                 lpParameter,
    0ul,      // _In_      DWORD                  dwCreationFlags,
    NULL      // _Out_opt_ LPDWORD                lpThreadId
  );

  if (!pEngine->thread.hThread) {
    DMESSAGE("creating thread");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineThread==pEngine->eState,exit);

  while (pEngine->eState<eEngineThreadRunning) {
    LeaveCriticalSection(&pEngine->cs); // {
    WaitForSingleObject(pEngine->thread.hConsumer,INFINITE);
    EnterCriticalSection(&pEngine->cs); // }
  }

  pEngine->thread.pCommand=NULL;
  LeaveCriticalSection(&pEngine->cs); // }

#if defined (YASAPI_ENGINE_ENUMERATOR) // [
  /////////////////////////////////////////////////////////////////////////////
  hr=CoCreateInstance(
    &CLSID_MMDeviceEnumerator,    // _In_  REFCLSID  rclsid,
    NULL,                         // _In_  LPUNKNOWN pUnkOuter,
    CLSCTX_ALL,                   // _In_  DWORD     dwClsContext,
    &IID_IMMDeviceEnumerator,     // _In_  REFIID    riid,
    (void **)&pEngine->pDeviceEnumerator
                                  // _Out_ LPVOID    *ppv
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("CoCreateInstance",REGDB_E_CLASSNOTREG);
    DCMESSAGE("CoCreateInstance",CLASS_E_NOAGGREGATION);
    DCMESSAGE("CoCreateInstance",E_NOINTERFACE);
    DCMESSAGE("CoCreateInstance",E_POINTER);
    DDMESSAGE("CoCreateInstance",hr);
    }

    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineDeviceEnumerator==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
  hr=pEngine->pDeviceEnumerator->lpVtbl->GetDefaultAudioEndpoint(
    pEngine->pDeviceEnumerator,
    eRender,                      // [in]  EDataFlow dataFlow,
    eConsole,                     // [in]  ERole     role,
    &pEngine->pDevice             // [out] IMMDevice **ppDevice
  );

#if defined (_MSC_VER) // [
  if (E_NOTFOUND==hr) {
    DMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint failed with"
        " E_NOTFOUND");
    EngineKill(pEngine,eState);
    goto exit;
  }
  else
#endif // ]
  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint",E_POINTER);
    DCMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint",E_INVALIDARG);
#if ! defined (_MSC_VER) // [
    DCMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint",E_NOTFOUND);
#endif // ]
    DCMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint",E_OUTOFMEMORY);
    DDMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint",hr);
    }

    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineDevice==pEngine->eState,exit);
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
exit:
  DWRITELN(EngineState(pEngine));
  return eState<pEngine->eState?0:-1;
}

static int EngineCreateOptions(Engine *pEngine, Player *pPlayer)
{
  int code=-1;
  OptionsCommon *pCommon=&pEngine->options.common;
  OptionsDevice *pDevice=&pEngine->options.device;
  Wasapi *pWasapi;

  /////////////////////////////////////////////////////////////////////////////
  if (!(pWasapi=PlayerWasapi(pPlayer))) {
    DMESSAGE("accessing WASAPI");
    EngineDestroy(pEngine);
    goto exit;
  }

  *pCommon=pWasapi->options.common.current;
  *pDevice=pWasapi->selected.pDevice->pMember->current;

  /////////////////////////////////////////////////////////////////////////////
  pEngine->params.pFriendlyName
      =pWasapi->selected.pDevice->pFriendlyNameOriginal;

  if (pDevice->eStrategy) {
    pEngine->params.dwStreamFlags=AUDCLNT_STREAMFLAGS_EVENTCALLBACK;
    DWRITELN("AUDCLNT_STREAMFLAGS_EVENTCALLBACK");
  }
  else
    pEngine->params.dwStreamFlags=0ul;

  if (pDevice->eMode) {
    pEngine->params.eShareMode=AUDCLNT_SHAREMODE_EXCLUSIVE;
    DWRITELN("AUDCLNT_SHAREMODE_EXCLUSIVE");
  }
  else {
    pEngine->params.eShareMode=AUDCLNT_SHAREMODE_SHARED;
    DWRITELN("AUDCLNT_SHAREMODE_SHARED");

    if (!pDevice->bAutoConvertPCM)
      goto shared;

    pEngine->params.dwStreamFlags|=AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM;
    DWRITELN("AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM");

    if (! pDevice->bSRCDefaultQuality)
      goto shared;

    pEngine->params.dwStreamFlags|=AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY;
    DWRITELN("AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY");
  shared: ;
  }

  DVWRITELN("dwStreamFlags=%0lx",pEngine->params.dwStreamFlags);
  pEngine->params.bDoubleBuffered=pDevice->eMode&&pDevice->eStrategy;

#if defined (YASAPI_OPTION_VOLUME) // [
  /////////////////////////////////////////////////////////////////////////////
  switch (pEngine->options.common.eVolume) {
  case eVolumeSoftware:
    pEngine->params.pIEngineVolumeVtbl=&gcIEngineVolumeSoftwareVtbl;
    break;
#if defined (YASAPI_VOLUME_WASAPI) // [
  case eVolumeWasapi:
    switch (pEngine->options.device.eMode) {
#if defined (YASAPI_SIMPLE_AUDIO_VOLUME) // [
    case eModeShared:
      pEngine->params.pIEngineVolumeVtbl=&gcIEngineVolumeSimpleAudioVtbl;
      break;
#endif // ]
#if defined (YASAPI_AUDIO_ENDPOINT_VOLUME) // [
    case eModeExclusive:
      pEngine->params.pIEngineVolumeVtbl=&gcIEngineVolumeAudioEndpointVtbl;
      break;
#endif // ]
    default:
      pEngine->params.pIEngineVolumeVtbl=&gcIEngineVolumeSoftwareVtbl;
      break;
    }

    break;
#endif // ]
  default:
    pEngine->params.pIEngineVolumeVtbl=&gcIEngineVolumeNullVtbl;
    break;
  };
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}

static int EngineCreateWfx(Engine *pEngine, WAVEFORMATEXTENSIBLE *pWfx,
    DWORD samplerate, WORD numchannels, WORD bitspersamp)
{
  enum {
    eSize=(sizeof *pWfx)-sizeof (pWfx->Format),
  };

  int code=-1;

  pWfx->Format.wFormatTag=WAVE_FORMAT_EXTENSIBLE;
  pWfx->Format.cbSize=eSize;
  pWfx->SubFormat=KSDATAFORMAT_SUBTYPE_PCM;

  switch (numchannels) {
  case 1:
    pWfx->dwChannelMask=KSAUDIO_SPEAKER_MONO;
    DWRITELN("KSAUDIO_SPEAKER_MONO");
    break;
  case 2:
    pWfx->dwChannelMask=KSAUDIO_SPEAKER_STEREO;
    DWRITELN("KSAUDIO_SPEAKER_STEREO");
    break;
  case 3:
    pWfx->dwChannelMask
        =SPEAKER_FRONT_LEFT
        |SPEAKER_FRONT_RIGHT
        |SPEAKER_LOW_FREQUENCY;
    DWRITELN("SPEAKER_FRONT_LEFT|SPEAKER_FRONT_RIGHT|SPEAKER_LOW_FREQUENCY");
    break;
  case 4:
    if (pEngine->options.common.bSurround) {
      pWfx->dwChannelMask=KSAUDIO_SPEAKER_SURROUND;
      DWRITELN("KSAUDIO_SPEAKER_SURROUND");
    }
    else {
      pWfx->dwChannelMask=KSAUDIO_SPEAKER_QUAD;
      DWRITELN("KSAUDIO_SPEAKER_QUAD");
    }

    break;
  case 5:
    pWfx->dwChannelMask
        =SPEAKER_BACK_RIGHT
        |SPEAKER_FRONT_RIGHT
        |SPEAKER_FRONT_CENTER
        |SPEAKER_FRONT_LEFT
        |SPEAKER_BACK_LEFT;
    DWRITELN("SPEAKER_BACK_RIGHT|SPEAKER_FRONT_RIGHT|SPEAKER_FRONT_CENTER"
        "|SPEAKER_FRONT_LEFT|SPEAKER_BACK_LEFT");
    break;
  case 6:
    if (pEngine->options.common.bSurround) {
      pWfx->dwChannelMask=KSAUDIO_SPEAKER_5POINT1_SURROUND;
      DWRITELN("KSAUDIO_SPEAKER_5POINT1_SURROUND");
    }
    else {
      pWfx->dwChannelMask=KSAUDIO_SPEAKER_5POINT1;
      DWRITELN("KSAUDIO_SPEAKER_5POINT1");
    }

    break;
  case 8:
    if (pEngine->options.common.bSurround) {
      pWfx->dwChannelMask=KSAUDIO_SPEAKER_7POINT1_SURROUND;
      DWRITELN("KSAUDIO_SPEAKER_7POINT1_SURROUND");
    }
    else {
      pWfx->dwChannelMask=KSAUDIO_SPEAKER_7POINT1;
      DWRITELN("KSAUDIO_SPEAKER_7POINT1");
    }

    break;
  default:
    DVMESSAGE("%d channels not supported",numchannels);
    goto exit;
  }

  ////
  pWfx->Format.nSamplesPerSec=samplerate;
  pWfx->Format.nChannels=numchannels;
  pWfx->Format.wBitsPerSample=bitspersamp;
  DVWRITELN("*** samplerate=%lu numchannels=%u bitspersamp=%u",
      samplerate,numchannels,bitspersamp);
  ////
  pWfx->Format.nBlockAlign=pWfx->Format.nChannels;
  pWfx->Format.nBlockAlign*=pWfx->Format.wBitsPerSample>>3;

#if 0 // [
  pEngine->wfx.target=*pWfx;
#endif // ]

  ////
  pWfx->Format.nAvgBytesPerSec=pWfx->Format.nSamplesPerSec;
  pWfx->Format.nAvgBytesPerSec*=pWfx->Format.nBlockAlign;
  ////
  pWfx->Samples.wValidBitsPerSample=pWfx->Format.wBitsPerSample;

  ////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}

static int EngineGetBufferDuration(Engine *pEngine)
{
  int code=-1;
  REFERENCE_TIME hnsBufferDuration;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
  if (pEngine->options.device.eMode) {
    hr=pEngine->pAudioClient->lpVtbl->GetDevicePeriod(pEngine->pAudioClient,
      NULL,       // [out] REFERENCE_TIME *phnsDefaultDevicePeriod,
      &hnsBufferDuration
                  // [out] REFERENCE_TIME *phnsMinimumDevicePeriod
    );
  }
  else {
    hr=pEngine->pAudioClient->lpVtbl->GetDevicePeriod(pEngine->pAudioClient,
      &hnsBufferDuration,
                  // [out] REFERENCE_TIME *phnsDefaultDevicePeriod,
      NULL        // [out] REFERENCE_TIME *phnsMinimumDevicePeriod
    );
  }

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClient::GetDevicePeriod",AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioClient::GetDevicePeriod",AUDCLNT_E_SERVICE_NOT_RUNNING);
    DCMESSAGE("IAudioClient::GetDevicePeriod",E_POINTER);
    DDMESSAGE("IAudioClient::GetDevicePeriod",hr);
    }

    goto exit;
  }

  DVWRITELN("*** hnsBufferDuration=%I64d ***",hnsBufferDuration);
  pEngine->params.hnsBufferDuration=hnsBufferDuration;

  /////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}

int EngineCreateAudioClient(Engine *pEngine, const EEngine eState)
{
  int code=-1;
  int nAlignment=0;
  UINT32 uNumFrames;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
activate:
  hr=pEngine->pDevice->lpVtbl->Activate(pEngine->pDevice,
    &IID_IAudioClient,      // [in]  REFIID      iid,
    CLSCTX_ALL,             // [in]  DWORD       dwClsCtx,
    NULL,                   // [in]  PROPVARIANT *pActivationParams,
    (void **)&pEngine->pAudioClient
                            // [out] void        **ppInterface
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IMMDevice::Activate",E_NOINTERFACE);
    DCMESSAGE("IMMDevice::Activate",E_POINTER);
    DCMESSAGE("IMMDevice::Activate",E_INVALIDARG);
    DCMESSAGE("IMMDevice::Activate",E_OUTOFMEMORY);
    DCMESSAGE("IMMDevice::Activate",AUDCLNT_E_DEVICE_INVALIDATED);
    DDMESSAGE("IMMDevice::Activate",hr);
    }

    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (0ll==pEngine->params.hnsBufferDuration) {
    if (eEngineAudioClientInititial!=eState) {
      DMESSAGE("internal");
      goto exit;
    }
    else if (EngineGetBufferDuration(pEngine)<0) {
      DMESSAGE("retrieving buffer duration");
      goto exit;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  if (pEngine->options.device.eStrategy) {
    hr=pEngine->pAudioClient->lpVtbl->Initialize(pEngine->pAudioClient,
      pEngine->params.eShareMode,
                            // [in]       AUDCLNT_SHAREMODE ShareMode,
      pEngine->params.dwStreamFlags,
                            // [in]       DWORD             StreamFlags,
      pEngine->params.hnsBufferDuration,
                            // [in]       REFERENCE_TIME    hnsBufferDuration,
      pEngine->params.hnsBufferDuration,
                            // [in]       REFERENCE_TIME    hnsPeriodicity,
      &pEngine->wfx.target.Format,
                            // [in] const WAVEFORMATEX      *pFormat,
      NULL                  // [in]       LPCGUID           AudioSessionGuid
    );
  }
  else {
    hr=pEngine->pAudioClient->lpVtbl->Initialize(pEngine->pAudioClient,
      pEngine->params.eShareMode,
                            // [in]       AUDCLNT_SHAREMODE ShareMode,
      pEngine->params.dwStreamFlags,
                            // [in]       DWORD             StreamFlags,
      pEngine->params.hnsBufferDuration,
                            // [in]       REFERENCE_TIME    hnsBufferDuration,
      0ull,                 // [in]       REFERENCE_TIME    hnsPeriodicity,
      &pEngine->wfx.target.Format,
                            // [in] const WAVEFORMATEX      *pFormat,
      NULL                  // [in]       LPCGUID           AudioSessionGuid
    );
  }

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClient::Initialize",CO_E_NOTINITIALIZED);
    DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_ALREADY_INITIALIZED);
    DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_WRONG_ENDPOINT_TYPE);
    case AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED:
      if (pEngine->options.device.bAlignmentDance&&!nAlignment) {
        // https://blogs.msdn.microsoft.com/matthew_van_eerde/2009/04/03/sample-wasapi-exclusive-mode-event-driven-playback-app-including-the-hd-audio-alignment-dance/
        hr=pEngine->pAudioClient->lpVtbl->GetBufferSize(pEngine->pAudioClient,
          &uNumFrames                             // UINT32 *pNumBufferFrames
        );

        pEngine->params.hnsBufferDuration=
            10.0*1000                             // (hns / ms) *
            *1000                                 // (ms / s) *
            *uNumFrames                           // frames /
            /pEngine->wfx.target.Format.nSamplesPerSec
                                                  // (frames / s)
            +0.5                                  // rounding
            ;

        DVWRITELN("*** uNumFrames (first pass)=%u ***",uNumFrames);
        DVWRITELN("*** hnsBufferDuration (first pass)=%I64d ***",
            pEngine->params.hnsBufferDuration);

        --pEngine->eState;
        DVWRITELN("--< %s",EngineState(pEngine));
        hr=pEngine->pAudioClient->lpVtbl->Release(pEngine->pAudioClient);
        pEngine->pAudioClient=NULL;
        ++nAlignment;
        goto activate;
      }

      DMESSAGE("IAudioClient::Initialize failed with"
          " AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED");
      break;
    DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_BUFFER_SIZE_ERROR);
    DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_CPUUSAGE_EXCEEDED);
    DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_DEVICE_IN_USE);
    //DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_ENDPOINT_CREATE_FAILED);
    case AUDCLNT_E_ENDPOINT_CREATE_FAILED:
      if (1==pEngine->wfx.target.Format.nChannels
          &&!pEngine->options.common.eMono2Stereo) {
        DMESSAGE("IAudioClient::Initialize failed with"
            " AUDCLNT_E_ENDPOINT_CREATE_FAILED"
            " (consider switching on mono to stereo convertion)");
        goto exit;
      }

      DMESSAGE("IAudioClient::Initialize failed with"
          " AUDCLNT_E_ENDPOINT_CREATE_FAILED");
      break;
    DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_INVALID_DEVICE_PERIOD);
    DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_UNSUPPORTED_FORMAT);
    DCMESSAGE("IAudioClient::Initialize",
        AUDCLNT_E_EXCLUSIVE_MODE_NOT_ALLOWED);
    DCMESSAGE("IAudioClient::Initialize",
        AUDCLNT_E_BUFDURATION_PERIOD_NOT_EQUAL);
    DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_SERVICE_NOT_RUNNING);
    DCMESSAGE("IAudioClient::Initialize",E_POINTER);
    DCMESSAGE("IAudioClient::Initialize",E_INVALIDARG);
    DCMESSAGE("IAudioClient::Initialize",E_OUTOFMEMORY);
    DDMESSAGE("IAudioClient::Initialize",hr);
    }

    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (0u==pEngine->params.uNumFramesDevice) {
    if (eEngineAudioClientInititial==eState) {
      hr=pEngine->pAudioClient->lpVtbl->GetBufferSize(pEngine->pAudioClient,
        &uNumFrames       // UINT32 *pNumBufferFrames
      );

      if (FAILED(hr)) {
        switch (hr) {
        DCMESSAGE("IAudioClient::GetBufferSize",AUDCLNT_E_NOT_INITIALIZED);
        DCMESSAGE("IAudioClient::GetBufferSize",AUDCLNT_E_DEVICE_INVALIDATED);
        DCMESSAGE("IAudioClient::GetBufferSize",AUDCLNT_E_SERVICE_NOT_RUNNING);
        DCMESSAGE("IAudioClient::GetBufferSize",E_POINTER);
        DDMESSAGE("IAudioClient::GetBufferSize",hr);
        }

        goto exit;
      }

      pEngine->params.uNumFramesDevice=uNumFrames;

      if (!pEngine->options.device.eStrategy) {
        pEngine->params.DueTime.QuadPart
            =-0.5*YASAPI_FRAMES2HNS(uNumFrames,&pEngine->wfx.target)-0.5;

        if (pEngine->options.device.bTimerPeriodic) {
          pEngine->params.lPeriod
              =0.5*YASAPI_FRAMES2MS(uNumFrames,&pEngine->wfx.target)+0.5;
        }
        else
          pEngine->params.lPeriod=0l;
      }
      else {
        pEngine->params.DueTime.QuadPart=0ll;
        pEngine->params.lPeriod=0l;
      }
    }
  }

//#if ! defined (YASAPI_ENGINE_REGISTER_WHEN_STARING) // [
  /////////////////////////////////////////////////////////////////////////////
  if (pEngine->options.device.eStrategy&&!pEngine->options.common.bDisconnect) {
    hr=pEngine->pAudioClient->lpVtbl->SetEventHandle(pEngine->pAudioClient,
      pEngine->timer.hEvent   // [in] HANDLE eventHandle
    );

    if (FAILED(hr)) {
      switch (hr) {
      DCMESSAGE("IAudioClient::SetEventHandle",E_INVALIDARG);
      DCMESSAGE("IAudioClient::SetEventHandle",
          AUDCLNT_E_EVENTHANDLE_NOT_EXPECTED);
      DCMESSAGE("IAudioClient::SetEventHandle",AUDCLNT_E_NOT_INITIALIZED);
      DCMESSAGE("IAudioClient::SetEventHandle",AUDCLNT_E_DEVICE_INVALIDATED);
      DCMESSAGE("IAudioClient::SetEventHandle",AUDCLNT_E_SERVICE_NOT_RUNNING);
      DDMESSAGE("IAudioClient::SetEventHandle",hr);
      }
  
      EngineKill(pEngine,eState);
      goto exit;
    }

    DWRITELN("event handle registered");
  }
//#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}

///////////////////////////////////////////////////////////////////////////////
void EngineKill(Engine *pEngine, const EEngine eState)
{
  Gtk *pGtk=&pEngine->pPlayer->gtk;

  switch (pEngine->eState) {
  case eEngineEot:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineEot==pEngine->eState,exit);
    --pEngine->eState;
    // Intentional fall-through.
  case eEngineDraining:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineDraining==pEngine->eState,exit);
    --pEngine->eState;
  case eEngineTimerStarted:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineTimerStarted==pEngine->eState,exit);
    --pEngine->eState;

    if (!pEngine->options.device.eStrategy)
      CancelWaitableTimer(pEngine->timer.hTimer);

    // Intentional fall-through.
  case eEngineAudioClientStarted:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineAudioClientStarted==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->pAudioClient->lpVtbl->Stop(pEngine->pAudioClient);

    if (pEngine->params.bDoubleBuffered)
      pEngine->pAudioClient->lpVtbl->Reset(pEngine->pAudioClient);
    // Intentional fall-through.
  case eEnginePauseConnected:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEnginePauseConnected==pEngine->eState,exit);
    --pEngine->eState;
    // Intentional fall-through.
  case eEngineAudioRenderClient:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineAudioRenderClient==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->pAudioRenderClient->lpVtbl->Release(pEngine->pAudioRenderClient);
    pEngine->pAudioRenderClient=NULL;
    // Intentional fall-through.
  case eEngineAudioClock:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineAudioClock==pEngine->eState,exit);

    if (!pEngine->options.common.bLatency) {
      pEngine->params.msHistory+=EngineOutputTimeInt(pEngine);
      pEngine->pAudioClock->lpVtbl->Release(pEngine->pAudioClock);
      pEngine->pAudioClock=NULL;
    }

    --pEngine->eState;
    // Intentional fall-through.
#if defined (YASAPI_OPTION_VOLUME) // [
  case eEngineVolume:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineVolume==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->params.pIEngineVolumeVtbl->Destroy(pEngine);
    // Intentional fall-through.
#endif // ]
  case eEngineAudioClient:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineAudioClient==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->pAudioClient->lpVtbl->Release(pEngine->pAudioClient);
    pEngine->pAudioClient=NULL;
    // Intentional fall-through.
  case eEnginePauseDisconnected:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEnginePauseDisconnected==pEngine->eState,exit);
    --pEngine->eState;
    // Intentional fall-through.
  case eEngineRing:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineRing==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->monitor.xWritten=0.0;
    pEngine->monitor.xPadding=0.0;
    RingDestroy(&pEngine->ring);

    ZeroMemory(&pEngine->params,sizeof pEngine->params);
    ZeroMemory(&pEngine->stats,sizeof pEngine->stats);
    pEngine->stats.uNumTracks=UINT32_MAX;
    pEngine->stats.uNumDropouts=UINT32_MAX;
    GtkUpdateDialog(pGtk,&pGtk->config);
    // Intentional fall-through.
  case eEngineAudioClientInititial:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineAudioClientInititial==pEngine->eState,exit);
    --pEngine->eState;

    if (pEngine->pAudioClient) {
      pEngine->pAudioClient->lpVtbl->Release(pEngine->pAudioClient);
      pEngine->pAudioClient=NULL;
    }
    // Intentional fall-through.
#if defined (YASAPI_ENGINE_ENUMERATOR) // [
  case eEngineDevice:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineDevice==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->pDevice->lpVtbl->Release(pEngine->pDevice);
    pEngine->pDevice=NULL;
    // Intentional fall-through.
  case eEngineDeviceEnumerator:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineDeviceEnumerator==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->pDeviceEnumerator->lpVtbl->Release(pEngine->pDeviceEnumerator);
    pEngine->pDeviceEnumerator=NULL;
    // Intentional fall-through.
#endif // ]
  case eEngineThreadRunning:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineThreadRunning==pEngine->eState,exit);
    --pEngine->eState;
    EngineSend(pEngine,eCommandExit,eEngineNull,1);
    WaitForSingleObject(pEngine->thread.hThread,INFINITE);
    // Intentional fall-through.
  case eEngineThread:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineThread==pEngine->eState,exit);
    --pEngine->eState;
    CloseHandle(pEngine->thread.hThread);
    pEngine->thread.hThread=NULL;
    // Intentional fall-through.
  case eEngineEventConsumer:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineEventConsumer==pEngine->eState,exit);
    --pEngine->eState;
    CloseHandle(pEngine->thread.hConsumer);
    pEngine->thread.hConsumer=NULL;
    // Intentional fall-through.
  case eEngineEventProducer:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineEventProducer==pEngine->eState,exit);
    --pEngine->eState;
    CloseHandle(pEngine->thread.hProducer);
    pEngine->thread.hProducer=NULL;
    // Intentional fall-through.
  case eEngineTimer:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineTimer==pEngine->eState,exit);
    --pEngine->eState;
    CloseHandle(pEngine->timer.hTimer);
    pEngine->timer.hTimer=NULL;
    // Intentional fall-through.
  case eEngineTimerEvent:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineTimerEvent==pEngine->eState,exit);
    --pEngine->eState;
    CloseHandle(pEngine->timer.hEvent);
    pEngine->timer.hEvent=NULL;
    // Intentional fall-through.
  case eEngineCriticalSection:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineCriticalSection==pEngine->eState,exit);
    --pEngine->eState;
    DeleteCriticalSection(&pEngine->cs);
    // Intentional fall-through.
  default:
    break;
  }

  DWRITELN(EngineState(pEngine));
#if defined (YASAPI_DEBUG) // [
exit:
  return;
#endif // ]
}

void EngineDestroy(Engine *pEngine)
{
  EngineKill(pEngine,eEngineNull);
}

void EngineClose(Engine *pEngine)
{
DMARKLN();
  EngineSend(pEngine,eCommandKill,eEngineBase,1);
DMARKLN();
}

///////////////////////////////////////////////////////////////////////////////
int EnginePause(Engine *pEngine, int nPause)
{
  int nLastPause=pEngine->params.nLastPause;

  if (nLastPause==nPause)
    goto epause;

  if (pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs); // {

  if (nPause) {
    switch (pEngine->eState) {
    case eEngineStarted:
      if (pEngine->options.common.bDisconnect)
        EngineSend(pEngine,eCommandKill,eEnginePauseDisconnected,0);
      else
        EngineSend(pEngine,eCommandKill,eEnginePauseConnected,0);

      pEngine->params.nLastPause=nPause;
      break;
    default:
      break;
    }
  }
  else {
    switch (pEngine->eState) {
    case eEnginePauseDisconnected:
    case eEnginePauseConnected:
      EngineSend(pEngine,eCommandStart,eEngineNull,0);
      pEngine->params.nLastPause=nPause;
      break;
    default:
      break;
    }
  }

  LeaveCriticalSection(&pEngine->cs); // }
ebase:
epause:
  return nLastPause;
}

int EngineIsPlaying(Engine *pEngine)
{
  int bIsPlaying=0;

  if (pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs); // {

  switch (pEngine->eState) {
  case eEnginePauseDisconnected:
  case eEnginePauseConnected:
    break;
  case eEngineStarted:
    ++pEngine->eState;
    DASSERT(eEngineDraining==pEngine->eState,exit);
    bIsPlaying=1;
    break;
  case eEngineDraining:
    bIsPlaying=1;
    break;
  default:
    break;
  }

#if defined (YASAPI_DEBUG) // [
exit:
#endif // ]
  LeaveCriticalSection(&pEngine->cs); // }
ebase:
DVWRITELN("%s (%d)",EngineState(pEngine),bIsPlaying);
  return bIsPlaying;
}

int EngineAvailable(Engine *pEngine)
{
  Ring *pRing=&pEngine->ring;
  int nAvailable=0;

  if (pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs); // {

  switch (pEngine->eState) {
  case eEnginePauseDisconnected:
  case eEnginePauseConnected:
    break;
  default:
    if (pEngine->eState<eEngineRing)
      nAvailable=INT_MAX;
    else
      nAvailable=pRing->lpVtbl->source.Available(pRing);

    break;
  }

  LeaveCriticalSection(&pEngine->cs); // }
ebase:
  return nAvailable;
}

#if defined (YASAPI_ENGINE_IS_PAUSE) // [
static int EngineIsPause(Engine *pEngine)
{
  int bIsPause=0;

  if (pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs); // {

  switch (pEngine->eState) {
  case eEnginePauseDisconnected:
  case eEnginePauseConnected:
    bIsPause=1;
  default:
    break;
  }
exit:
  LeaveCriticalSection(&pEngine->cs); // {
ebase:
  return bIsPause;
}
#endif // ]

static double EngineOutputTimeInt(Engine *pEngine)
{
  double msOutputTime=0.0;
  UINT64 u64Frequency;
  UINT64 u64Position;
  HRESULT hr;

  if (pEngine->options.common.bLatency||pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs);

#if defined (YASAPI_ENGINE_IS_PAUSE) // [
  if (EngineIsPause(pEngine))
    goto exit;
#endif // ]

  if (pEngine->eState<eEngineAudioClock)
    goto exit;

  if (eEnginePauseDisconnected==pEngine->eState)
    goto exit;

  hr=pEngine->pAudioClock->lpVtbl->GetFrequency(pEngine->pAudioClock,
    &u64Frequency     // [out] UINT64 *pu64Frequency
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClock::GetFrequency",E_POINTER);
    DCMESSAGE("IAudioClock::GetFrequency",AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioClock::GetFrequency",AUDCLNT_E_SERVICE_NOT_RUNNING);
    DDMESSAGE("IAudioClock::GetFrequency",hr);
    }

    goto exit;
  }

  hr=pEngine->pAudioClock->lpVtbl->GetPosition(pEngine->pAudioClock,
    &u64Position,     // [out] UINT64 *pu64Position
    NULL              // UINT64 *pu64QPCPosition
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClock::GetPosition",E_POINTER);
    DCMESSAGE("IAudioClock::GetPosition",AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioClock::GetPosition",AUDCLNT_E_SERVICE_NOT_RUNNING);
    DDMESSAGE("IAudioClock::GetPosition",hr);
    }

    goto exit;
  }

  msOutputTime+=1000.0*u64Position/u64Frequency;
exit:
  LeaveCriticalSection(&pEngine->cs);
ebase:
  return msOutputTime;
}

int EngineOutputTimeExt(Engine *pEngine)
{
#if defined (YASAPI_ENGINE_IS_PAUSE) // [
  if (EngineIsPause(pEngine))
    return 0;
  else {
#endif // ]
    return EngineOutputTimeInt(pEngine)-pEngine->params.msHistory+0.5;
#if defined (YASAPI_ENGINE_IS_PAUSE) // [
  }
#endif // ]
}

int EngineWrittenTime(Engine *pEngine)
{
  DWORD nAvgBytesPerSec=pEngine->wfx.target.Format.nAvgBytesPerSec;
  double msWrittenTime=0.0;
  Ring *pRing=&pEngine->ring;
  DWORD dwWritten;
  UINT32 uNumFramesPadding;

  if (pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs); // {

#if defined (YASAPI_ENGINE_IS_PAUSE) // [
  if (EngineIsPause(pEngine))
    goto exit;
#endif // ]

  if (pEngine->eState<eEngineRing)
    goto exit;

  dwWritten=pRing->lpVtbl->target.Written(pRing);
  msWrittenTime+=MulDiv(1000,dwWritten,nAvgBytesPerSec);
  msWrittenTime+=EngineOutputTimeInt(pEngine);
  msWrittenTime-=pEngine->params.msHistory;

  if (!pEngine->options.common.bLatency&&!pEngine->params.bDoubleBuffered) {
    if (EngineNumFramesPadding(pEngine,&uNumFramesPadding)<0)
      goto exit;

    msWrittenTime+=YASAPI_FRAMES2MS(uNumFramesPadding,&pEngine->wfx.target);
  }
exit:
  LeaveCriticalSection(&pEngine->cs); // }
ebase:
  return msWrittenTime+0.5;
}

static void EngineRoundUp(DWORD *pdwSize, DWORD dwChunkSizeSource)
{
  DWORD dwMod=*pdwSize%dwChunkSizeSource;

  if (dwMod)
    *pdwSize+=dwChunkSizeSource-dwMod;
}

#if defined (YASAPI_DEBUG) // [
// https://stackoverflow.com/questions/19738919/gcd-function-for-c
UINT32 gcd(UINT32 a, UINT32 b)
{
  while (b) {
    a%=b;
    a^=b;
    b^=a;
    a^=b;
  }

  return a;
}

// https://stackoverflow.com/questions/19738919/gcd-function-for-c
UINT32 lcm(UINT32 a, UINT32 b)
{
  return MulDiv(a,b,gcd(a,b));
}
#endif // ]

static int EngineCreateRing(Engine *pEngine, DWORD dwChunkSizeSource)
{
  int code=-1;
  Ring *pRing=&pEngine->ring;
  Gtk *pGtk=&pEngine->pPlayer->gtk;
  WORD nBlockAlignSource=pEngine->wfx.source.Format.nBlockAlign;
  EEngine eState=pEngine->eState;
  double xPlaybackBegin=pEngine->options.device.xPlaybackBegin;
  UINT32 uNumFramesDevice=pEngine->params.uNumFramesDevice;
  WAVEFORMATEXTENSIBLE *pWfxTarget=&pEngine->wfx.target;
  WAVEFORMATEXTENSIBLE *pWfxSource=&pEngine->wfx.source;
  DWORD dwSizeSharedSource;
  DWORD dwPlaybackBeginSource;
  DWORD dwSizeRingSource;
  IRingVtbl *lpRingVtbl;
#if defined (YASAPI_OPTION_VOLUME) // [
  int bFormatTransform;
#endif // ]
#if 0 && defined (YASAPI_DEBUG) // [
  UINT32 uChunkSize=dwChunkSizeSource/nBlockAlign;
  UINT32 uLcm=lcm(pEngine->params.uNumFramesDevice,uChunkSize);

  DVWRITELN("uNumFramesDevice=%u uChunkSize=%u uLcm=%u %u %u",
      pEngine->params.uNumFramesDevice,uChunkSize,uLcm,
          uLcm/pEngine->params.uNumFramesDevice,uLcm/uChunkSize);
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  // We round up dwSizeRing/dwPlaybackBegin to a multiple of len in order that
  // multiple chunks fit.
  dwSizeSharedSource=uNumFramesDevice*nBlockAlignSource;

  dwPlaybackBeginSource=xPlaybackBegin*dwSizeSharedSource+0.5;
  EngineRoundUp(&dwPlaybackBeginSource,dwChunkSizeSource);
  pEngine->params.source.dwPlaybackBegin=dwPlaybackBeginSource;

  dwSizeRingSource=pEngine->options.device.xRingSize*dwChunkSizeSource+0.5;

  // We make sure that at least two chunks fit into the buffer.
  if (dwSizeRingSource<dwPlaybackBeginSource+(dwChunkSizeSource<<1))
    dwSizeRingSource=dwPlaybackBeginSource+(dwChunkSizeSource<<1);

  pEngine->params.uNumFramesRing=dwSizeRingSource/nBlockAlignSource;
  DVWRITELN("dwPlaybackBegin=%lu dwSizeRing=%lu",
      dwPlaybackBeginSource/nBlockAlignSource,
      dwSizeRingSource/nBlockAlignSource);

  GtkUpdateWidget(pGtk,&pGtk->config.config.monitor.engine.num_frames_ring);

  /////////////////////////////////////////////////////////////////////////////
  // The ring buffer is set up wrt to the target format.
#if defined (YASAPI_OPTION_VOLUME) // [
  bFormatTransform=memcmp(&pWfxTarget->Format,&pWfxSource->Format,
      sizeof pWfxTarget->Format);

  if (eVolumeSoftware==pEngine->options.common.eVolume||bFormatTransform) {
#else // ] [
  if (memcmp(pWfxTarget,pWfxSource,sizeof *pWfxTarget)) {
#endif // ]
    lpRingVtbl=&gcIRingTransformVtbl;
#if ! defined (YASAPI_OPTION_VOLUME) // [
  }
#else // ] [
  }
#endif // ]
  else
    lpRingVtbl=&gcIRingPlainVtbl;

  if (RingCreate(pRing,lpRingVtbl,pEngine,dwSizeRingSource)<0) {
    DMESSAGE("creating ring");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DWRITELN(EngineState(pEngine));
  DASSERT(eEngineRing==pEngine->eState,exit);
  code=0;
exit:
  DWRITELN(EngineState(pEngine));
  return code;
}

#define SIZE_TARGET
int EngineWrite(Engine *pEngine, BYTE *buf, DWORD dwChunkSize)
{
  int code=-1;
  Ring *pRing=&pEngine->ring;
  Player *pPlayer=pEngine->pPlayer;
  Gtk *pGtk=&pPlayer->gtk;
#if defined (SIZE_TARGET) // [
  DWORD dwSizeTarget;
#else // ] [
  DWORD dwSizeSource;
#endif // ]
  DWORD dwWrittenSource;
  DWORD dwWrittenTarget;
  UINT32 uNumFrames;
  UINT32 uMsLatency;

  if (pEngine->eState<eEngineBase) {
    DMESSAGE("engine not initialized");
    goto ebase;
  }

  EnterCriticalSection(&pEngine->cs); // {

  switch (pEngine->eState) {
  case eEnginePauseDisconnected:
  case eEnginePauseConnected:
    goto exit;
  default:
    break;
  }

  if (pEngine->eState<eEngineRing&&EngineCreateRing(pEngine,dwChunkSize)<0) {
    DMESSAGE("creating ring");
    goto exit;
  }

#if 0 // [
DVWRITELN("dwChunkSize=%lu {",dwChunkSize);
  dwChunkSize=pRing->lpVtbl->target.ConvertSize(pRing,dwChunkSize);
DVWRITELN("dwChunkSize=%lu }",dwChunkSize);
#endif // ]

  if (RingWrite(&pEngine->ring,buf,dwChunkSize)<0) {
    DMESSAGE("writing to ring");
    goto exit;
  }

#if defined (SIZE_TARGET) // [
  dwSizeTarget=pRing->lpVtbl->target.Size(pRing);
#else // ] [
  dwSizeSource=pRing->lpVtbl->source.Size(pRing);
#endif // ]
  dwWrittenSource=pRing->lpVtbl->source.Written(pRing);
  dwWrittenTarget=pRing->lpVtbl->target.Written(pRing);

  uNumFrames=YASAPI_SIZE2FRAMES(dwWrittenTarget,&pEngine->wfx.target);
  uMsLatency=YASAPI_FRAMES2MS(uNumFrames,&pEngine->wfx.target)+0.5;

  if (pPlayer->stats.uMsLatencyMax<uMsLatency) {
    pPlayer->stats.uMsLatencyMax=uMsLatency;
    GtkUpdateWidget(pGtk,
        &pGtk->config.config.monitor.player.ms_latency);
  }

  if (pEngine->stats.uNumFramesMax<uNumFrames) {
    pEngine->stats.uNumFramesMax=uNumFrames;
    GtkUpdateWidget(pGtk,
        &pGtk->config.config.monitor.engine.num_frames_latency);
  }

  if (eEngineStarted<=pEngine->eState)
    goto success;

  if (dwWrittenSource<pEngine->params.source.dwPlaybackBegin) {
    DVWRITELN("*** undeflow: %lu (%lu)",
        dwWrittenSource/pEngine->wfx.source.Format.nBlockAlign,
        pEngine->params.source.dwPlaybackBegin
            /pEngine->wfx.source.Format.nBlockAlign);
#if defined (SIZE_TARGET) // [
    pEngine->monitor.xWritten=(double)dwWrittenTarget/dwSizeTarget;
#else // ] [
    pEngine->monitor.xWritten=(double)dwWrittenSource/dwSizeSource;
#endif // ]
    GtkUpdateWidget(pGtk,&pGtk->config.config.monitor.written);
    goto success;
  }

  DVWRITELN("*** starting: %lu (%lu)",
      dwWrittenSource/pEngine->wfx.source.Format.nBlockAlign,
      pEngine->params.source.dwPlaybackBegin
          /pEngine->wfx.source.Format.nBlockAlign);

  EngineSend(pEngine,eCommandStart,eEngineNull,0);
  DWRITELN(EngineState(pEngine));
success:
  code=0;
exit:
  LeaveCriticalSection(&pEngine->cs); // }
ebase:
  return code;
}

#if defined (YASAPI_OPTION_VOLUME) // [
void EngineVolume(Engine *pEngine)
{
  if (pEngine->eState<eEngineBase) {
    DWRITELN("engine not initialized");
    goto ebase;
  }

  EnterCriticalSection(&pEngine->cs); // {

  if (pEngine->eState<eEngineStarted) {
    DWRITELN(EngineState(pEngine));
    goto exit;
  }

  pEngine->params.pIEngineVolumeVtbl->Set(pEngine);
exit:
  LeaveCriticalSection(&pEngine->cs); // }
ebase:
  return;
}
#endif // ]

///////////////////////////////////////////////////////////////////////////////
void EngineSend(Engine *pEngine, const ECommand eCommand, EEngine eArgument,
  int bLock)
{
  Command command={ eCommand, eArgument };

  if (bLock)
    EnterCriticalSection(&pEngine->cs); // {

  pEngine->thread.pCommand=&command;
  SetEvent(pEngine->thread.hProducer);

  while (pEngine->thread.pCommand) {
    LeaveCriticalSection(&pEngine->cs); // {
    WaitForSingleObject(pEngine->thread.hConsumer,INFINITE);
    EnterCriticalSection(&pEngine->cs); // }
  }

  if (bLock)
    LeaveCriticalSection(&pEngine->cs); // }
}

#if defined (YASAPI_DEBUG) // [
///////////////////////////////////////////////////////////////////////////////
const char *EngineStateEx(EEngine eState)
{
  const char *mnemonic[1+eEngineMax]={
    "undefined",
    "eEngineNull",
    "eEngineCriticalSection",
    "eEngineTimerEvent",
    "eEngineTimer",
    "eEngineEventProducer",
    "eEngineEventConsumer",
    "eEngineThread",
    "eEngineThreadRunning",
#if defined (YASAPI_ENGINE_ENUMERATOR) // [
    "eEngineDeviceEnumerator",
    "eEngineDevice",
#endif // ]
    "eEngineAudioClientInititial",
    "eEngineRing",
    "eEnginePauseDisconnected",
    "eEngineAudioClient",
#if defined (YASAPI_OPTION_VOLUME) // [
    "eEngineVolume",
#endif // ]
    "eEngineAudioClock",
    "eEngineAudioRenderClient",
    "eEnginePauseConnected",
    "eEngineAudioClientStarted",
    "eEngineTimerStarted",
    "eEngineDraining",
    "eEngineEot",
  };

  return mnemonic[0<=eState&&eState<eEngineMax?1+eState:0];
}

const char *EngineState(Engine *pEngine)
{
  return EngineStateEx(pEngine->eState);
}

const char *EngineCommandEx(ECommand eCommand)
{
  const char *mnemonic[1+eCommandMax]={
    "undefined",
    "eCommandNull",
    "eCommandExit",
    "eCommandStart",
    "eCommandFlush",
    "eCommandKill",
  };

  return mnemonic[0<=eCommand&&eCommand<eCommandMax?1+eCommand:0];
}

const char *EngineCommand(Engine *pEngine)
{
  Command *pCommand=pEngine->thread.pCommand;

  return pCommand?EngineCommandEx(pCommand->eCommand):"NULL";
}
#endif // ]
