/*
 * yasapi_ring.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

int RingCreate(Ring *pRing, IRingVtbl *lpVtbl, Engine *pEngine, DWORD dwSize)
{
  pRing->lpVtbl=lpVtbl;
  pRing->pEngine=pEngine;

  dwSize=lpVtbl->target.ConvertSize(pRing,dwSize);

  if (!(pRing->_buf=malloc(dwSize))) {
    DMESSAGE("allocation ring");
    goto emalloc;
  }

  pRing->_rp=pRing->_buf;
  pRing->_wp=pRing->_buf;
  pRing->_mp=pRing->_buf+dwSize;
  pRing->_dwWritten=0ul;
  pRing->_dwAvailable=dwSize;
  DVWRITELN("dwAvailable=%lu dwWritten=%lu",pRing->_dwAvailable,
      pRing->_dwWritten);
emalloc:
  return 0;
}

void RingDestroy(Ring *pRing)
{
  if (pRing->_buf) {
    free(pRing->_buf);
    pRing->_buf=NULL;
  }

  pRing->pEngine=NULL;
}

void RingReset(Ring *pRing)
{
  if (!pRing->pEngine) {
    DMESSAGE("ring not initialized");
    goto enoinit;
  }

  EnterCriticalSection(&pRing->pEngine->cs); // {
  pRing->_rp=pRing->_buf;
  pRing->_wp=pRing->_buf;
  pRing->_dwWritten=0ul;
  pRing->_dwAvailable=pRing->_mp-pRing->_buf;
  LeaveCriticalSection(&pRing->pEngine->cs); // }
enoinit:
  return;
}

int RingWrite(Ring *pRing, BYTE *buf, DWORD dwSize)
{
  int code=-1;
  DWORD dwSize1,dwSize2;

  if (!pRing->pEngine) {
    DMESSAGE("ring not initialized");
    goto enoinit;
  }

  EnterCriticalSection(&pRing->pEngine->cs); // {
  dwSize=pRing->lpVtbl->target.ConvertSize(pRing,dwSize);

  if (pRing->_wp<pRing->_rp) {
    // already wrapped
    if (pRing->_rp<pRing->_wp+dwSize) {
      DMESSAGE("overflow");
      goto eoverflow;
    }

    pRing->lpVtbl->Buf2Ring(pRing,pRing->_wp,buf,dwSize);
    pRing->_wp+=dwSize;
  }
  else if (pRing->_mp<pRing->_wp+dwSize) {
    // will wrap
    dwSize1=pRing->_mp-pRing->_wp;
    dwSize2=dwSize-dwSize1;

    if (pRing->_rp<pRing->_buf+dwSize2) {
      DMESSAGE("overflow");
      goto eoverflow;
    }

    pRing->lpVtbl->Buf2Ring(pRing,pRing->_wp,buf,dwSize1);
    pRing->lpVtbl->Buf2Ring(pRing,pRing->_buf,buf+dwSize1,dwSize2);
    pRing->_wp=pRing->_buf+dwSize2;
  }
  else {
    // will not wrap
    if (pRing->_mp<pRing->_wp+dwSize) {
      DMESSAGE("overflow");
      goto eoverflow;
    }

    pRing->lpVtbl->Buf2Ring(pRing,pRing->_wp,buf,dwSize);
    pRing->_wp+=dwSize;
  }

  pRing->_dwAvailable-=dwSize;
  pRing->_dwWritten+=dwSize;

  code=0;
eoverflow:
  LeaveCriticalSection(&pRing->pEngine->cs); // }
enoinit:
  return code;
}

#if defined (YASAPI_OPTION_VOLUME) // [
static void Ring2Buf(Ring *pRing, BYTE *pDst, const BYTE *pSrc,
    DWORD dwSize)
{
  Engine *pEngine=pRing->pEngine;
  IEngineVolumeVtbl *pIEngineVolumeVtbl=pEngine->params.pIEngineVolumeVtbl;
  int bApplyGain=pIEngineVolumeVtbl->bApplyGain;
  int nVolume;
  WORD wBytesPerSample;
  int32_t sample;
  BYTE *pSampleMin;
  const BYTE *pSrcMax;

  bApplyGain=bApplyGain&&(0<=(nVolume=pEngine->pPlayer->params.nVolume));
  bApplyGain=bApplyGain&&nVolume<ePlayerVolumeMax;

  if (bApplyGain) {
    wBytesPerSample=pEngine->wfx.target.Format.wBitsPerSample>>3;
    pSampleMin=(BYTE *)&sample+((sizeof sample)-wBytesPerSample);
    pSrcMax=pSrc+dwSize;

    while (pSrc<pSrcMax) {
      sample=0;
      memcpy(pSampleMin,pSrc,wBytesPerSample);
      pSrc+=wBytesPerSample;

      sample=MulDiv(sample,nVolume,ePlayerVolumeMax);
      memcpy(pDst,pSampleMin,wBytesPerSample);
      pDst+=wBytesPerSample;
    }
  }
  else
    memcpy(pDst,pSrc,dwSize);
}
#endif // ]

int RingRead(Ring *pRing, BYTE *buf, DWORD dwSize)
{
  int code=-1;
  DWORD dwSize1,dwSize2;

  if (!pRing->pEngine) {
    DMESSAGE("ring not initialized");
    goto enoinit;
  }

  EnterCriticalSection(&pRing->pEngine->cs); // {

  if (pRing->_rp<pRing->_wp) {
    // not wrapped
    if (pRing->_wp<pRing->_rp+dwSize) {
      DMESSAGE("underflow");
      goto eunderflow;
    }

#if defined (YASAPI_OPTION_VOLUME) // [
    Ring2Buf(pRing,buf,pRing->_rp,dwSize);
#else // ] [
    memcpy(buf,pRing->_rp,dwSize);
#endif // ]
    pRing->_rp+=dwSize;
  }
  else {
    // wrapped
    dwSize1=pRing->_mp-pRing->_rp;

    if (dwSize<dwSize1) {
#if defined (YASAPI_OPTION_VOLUME) // [
      Ring2Buf(pRing,buf,pRing->_rp,dwSize);
#else // ] [
      memcpy(buf,pRing->_rp,dwSize);
#endif // ]
      pRing->_rp+=dwSize;
    }
    else {
      dwSize2=dwSize-dwSize1;
#if defined (YASAPI_OPTION_VOLUME) // [
      Ring2Buf(pRing,buf,pRing->_rp,dwSize1);
      Ring2Buf(pRing,buf+dwSize1,pRing->_buf,dwSize2);
#else // ] [
      memcpy(buf,pRing->_rp,dwSize1);
      memcpy(buf+dwSize1,pRing->_buf,dwSize2);
#endif // ]
      pRing->_rp=pRing->_buf+dwSize2;
    }
  }

  pRing->_dwWritten-=dwSize;
  pRing->_dwAvailable+=dwSize;

  code=0;
eunderflow:
  LeaveCriticalSection(&pRing->pEngine->cs); // }
enoinit:
  return code;
}
